#pragma once

#include <QDoubleSpinBox>

/**
 * \brief Regular QDoubleSpinBox with some improvements.
 *
 * Changes:
 * - When using the arrows to step up or down, we only step to integers.
 * - We enable overwriting numbers behind the decimal point if the spin box is
 * full.
 */
class CustomDoubleSpinBox : public QDoubleSpinBox {
    Q_OBJECT
public:
    /**
     * \brief No changes to the QDoubleSpinBox-constructor.
     *
     * Constructs a spin box with 0.0 as minimum value and 99.99 as maximum
     * value, a step value of 1.0 and a precision of 2 decimal places. The value
     * is initially set to 0.00.
     *
     * \param parent Parent of the spin box.
     */
    explicit CustomDoubleSpinBox(QWidget *parent = nullptr);

    /**
     * \brief Implements rounding when changing the value via up and down
     * buttons.
     *
     * If the current value is a multiple of singleStep, add or subtract the
     * given number of steps. Else, round in the selected direction to the next
     * multiple of singleStep and then move (steps-1) steps.
     *
     * \param steps How many steps are supposed to be taken.
     */
    void stepBy(int steps) override;

    /**
     * \brief Implements overwriting of numbers behind the decimal point.
     *
     * The default behaviour doesn't allow you to write in this situation
     * (decimals = 2): 20.|00 (| being the cursor). This override allows the
     * user to overwrite numbers behind the decimal point.
     *
     * If there are is still space for more decimals, then nothing gets
     * overwritten.
     *
     * \note Currently 20.|30 -> press 6 -> 20.60, which is not ideal.
     *
     * \param event Pressed key.
     */
    void keyPressEvent(QKeyEvent *event) override;
};
