#include "export_rwthonline.hpp"

#include "data/csv_dump.hpp"
#include "data/data_manager.hpp"
#include "file_io/csv_parser.hpp"
#include "file_io/file.hpp"
#include "util/enums.hpp"
#include "util/grade.hpp"
#include "util/points.hpp"
#include "util/strings.hpp"

#include <QFileDialog>
#include <QString>
#include <QTableWidget>

#include <cstdio>
#include <ranges>

namespace {
qsizetype getColumn(const CsvLine &headerLine, const QString &headerText) {

    for (qsizetype i = 0; i < headerLine.size(); ++i) {

        if (QString::compare(headerLine[i], headerText, Qt::CaseInsensitive) ==
            0) {
            return i;
        }
    }
    return headerLine.size();
}

[[nodiscard]] QString getRemark(const GradingType type, const qsizetype idx) {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return "";

    const auto nrDecPlaces =
        nrDecimalPlaces(cur->mUserPointConfig.mGradeBoundaryStepSize);
    const auto pointsWithoutBonus = QString::number(
        cur->getPointsWithoutBonus(type)[idx].p, 'f', nrDecPlaces);
    const auto grade = cur->getGradeList(type)[idx];
    const auto nrBonusPoints =
        QString::number(cur->mBonus.mTotalBonusPointsPerStudent
                            .value(cur->mExam.mIdentifier[idx])
                            .p,
                        'f', nrDecPlaces);
    const auto maxGradeBoundary =
        type == GradingType::Closed
            ? cur->mChangedMaxPointsClosed.mChangedMaxPointBoundary
            : cur->mChangedMaxPointsOpen.mChangedMaxPointBoundary;
    const auto max = QString::number(
        toPoints(maxGradeBoundary, cur->mUserPointConfig.mGradeBoundaryStepSize)
            .p,
        'f', nrDecPlaces);

    QString remark;
    if (cur->hasBonus()) {
        // Change position of bonus points in the remark depending on whether or
        // not the student passed. This is to clarify that passing due to bonus
        // points might not possible.
        const auto remarkTemplate =
            grade == Grade::grade5_0
                ? str::fileExport::rwthOnlineRemarkNotPassedBonus
                : str::fileExport::rwthOnlineRemarkPassedBonus;
        remark = QString(remarkTemplate)
                     .arg(toString(type))
                     .arg(pointsWithoutBonus)
                     .arg(toString(grade))
                     .arg(nrBonusPoints)
                     .arg(max);
    } else {
        remark = QString(str::fileExport::rwthOnlineRemarkNoBonus)
                     .arg(toString(type))
                     .arg(pointsWithoutBonus)
                     .arg(toString(grade))
                     .arg(max);
    }

    if (type == GradingType::Closed) {
        // Add info on average points of candidates taking the exam for the
        // first time
        if (cur->mFirstTry.mFirstTryClosedPoints.empty()) {
            remark += str::fileExport::rwthOnlineRemarkAverageFirstNone;
        } else {
            remark += QString(str::fileExport::rwthOnlineRemarkAverageFirstAny)
                          .arg(cur->mFirstTry.mAvgFirstTryPoints.p);
        }
    }

    return remark;
}
} // namespace

void exportKnownStudents(QWidget *parent) {

    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    const auto exportLocationString = QFileDialog::getSaveFileName(
        parent, str::fileExport::knownStudentSaveLocationCaption, "",
        str::ui::input::csvFilter);

    if (!exportLocationString.isEmpty()) {

        auto exportCsvDump = cur->mStudents.mRwthOnline;
        auto &exportCsv = exportCsvDump.data;

        if (exportCsv.empty())
            return;

        // Insert student data into csv
        {
            const auto gradeColumn = getColumn(exportCsv[0], "grade");
            const auto remarkColumn = getColumn(exportCsv[0], "remark");
            if (gradeColumn >= exportCsv[0].size() ||
                remarkColumn >= exportCsv[0].size())
                return;

            // Clear previous cell content
            for (auto &l : exportCsv | std::views::drop(1)) {

                l[gradeColumn] = "";
                l[remarkColumn] = "";
            }

            // Insert new content where we have data
            for (qsizetype i = 0; auto &l : exportCsv | std::views::drop(1)) {

                using enum GradingType;

                // Search for same student in dynexite data
                const auto idx = cur->mExam.mIdentifier.indexOf(
                    cur->mStudents.mRegistrationNr[i]);

                // Convert to index, or skip if not found
                if (idx == -1) {
                    ++i;
                    continue;
                }

                // Save result in new csv
                l[gradeColumn] =
                    toString(cur->getGradeList(Combined)[idx], ',');
                // Only print closed/open points if there were any to earn
                if (cur->getMaxPoints(Closed) > Points()) {
                    l[remarkColumn] = getRemark(Closed, idx);
                }
                if (cur->getMaxPoints(Closed) > Points() &&
                    cur->getMaxPoints(Open) > Points()) {
                    l[remarkColumn] += " - ";
                }
                if (cur->getMaxPoints(Open) > Points()) {
                    l[remarkColumn] += getRemark(Open, idx);
                }

                ++i;
            }

            // Mark students without data as missing
            for (auto &l : exportCsv | std::views::drop(1)) {

                if (l[gradeColumn].isEmpty()) {
                    l[gradeColumn] = "X";
                }
            }
        }

        auto [file, stream] = openFileToWrite(exportLocationString);
        writeCsv(*stream, exportCsvDump);
    }
}

void exportUnknownStudents(QWidget *parent) {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    const auto exportLocationString = QFileDialog::getSaveFileName(
        parent, str::fileExport::unknownStudentSaveLocationCaption, "",
        str::ui::input::csvFilter);

    if (!exportLocationString.isEmpty()) {
        CsvDump exportDump;
        auto &exportCsv = exportDump.data;

        // Header line
        exportCsv = {{"REGISTRATION_NUMBER", "GRADE", "REMARK"}};

        // Insert students not in the input RWTHonline file
        for (qsizetype i = 0; auto &id : cur->mExam.mIdentifier) {

            using enum GradingType;

            // Search for same student in RWTHonline
            const auto pos = cur->mStudents.mRegistrationNr.indexOf(id);

            // Skip if found
            if (pos != -1) {
                ++i;
                continue;
            }

            // Save result in new line
            exportCsv.emplace_back(3);
            exportCsv.back()[0] = id;
            exportCsv.back()[1] = toString(cur->getGradeList(Combined)[i], ',');
            // Only print closed/open points if there were any to earn
            if (cur->getMaxPoints(Closed) > Points()) {
                exportCsv.back()[2] = getRemark(Closed, i);
            }
            if (cur->getMaxPoints(Closed) > Points() &&
                cur->getMaxPoints(Open) > Points()) {
                exportCsv.back()[2] += " - ";
            }
            if (cur->getMaxPoints(Open) > Points()) {
                exportCsv.back()[2] += getRemark(Open, i);
            }

            ++i;
        }

        auto [file, stream] = openFileToWrite(exportLocationString);
        writeCsv(*stream, exportDump);
    }
}

void exportManualAllStudents(QWidget *parent,
                             const QTableWidget &studentTable) {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    const auto exportLocationString = QFileDialog::getSaveFileName(
        parent, str::fileExport::allStudentsSaveLocationCaption, "",
        str::ui::input::csvFilter);

    if (!exportLocationString.isEmpty()) {
        // Store data from studentTable in a CsvDump
        CsvDump exportDump;
        auto &exportCsv = exportDump.data;

        const auto columns = studentTable.columnCount();
        const auto rows = studentTable.rowCount();

        // headers
        exportCsv.emplace_back();
        for (int column = 0; column < columns; ++column) {
            exportCsv.back().emplace_back(
                studentTable.horizontalHeaderItem(column)->text());
        }

        // data
        for (int row = 0; row < rows; ++row) {
            exportCsv.emplace_back();
            for (int column = 0; column < columns; ++column) {
                exportCsv.back().emplace_back(
                    studentTable.item(row, column)->text());
            }
        }

        auto [file, stream] = openFileToWrite(exportLocationString);
        writeCsv(*stream, exportDump);
    }
}
