#include "calculation_window.hpp"

#include "data/data_manager.hpp"

CalculationWindow::CalculationWindow(QWidget *parent)
    : QWidget(parent), mLayout(this) {

    setWindowTitle("Calculating the Closed Pass Mark");

    int currentRow = 0;

    int maxClosedRow;
    int avgFirstTryRow;
    int absPassMarkRow;
    int relPassMark0_5Row;
    int relPassMark0_78Row;

    {
        // Suppositions
        mLayout.addWidget(new QLabel("<b>Suppositions:</b>", this),
                          currentRow++, 0);
        maxClosedRow = currentRow;
        mLayout.addWidget(
            new QLabel("Maximum number of points that can be achieved:", this),
            currentRow++, 0);
        avgFirstTryRow = currentRow;
        mLayout.addWidget(
            new QLabel("Average number of points achieved by all "
                       "students taking the\nexam for the first time:",
                       this),
            currentRow++, 0);
    }

    {
        // Separator line
        const auto line = new QFrame(this);
        line->setFrameShape(QFrame::HLine);
        mLayout.addWidget(line, currentRow++, 0, 1, 3);
    }

    {
        // Absolute Pass Mark
        mLayout.addWidget(new QLabel("<b>Absolute Pass Mark:</b>", this),
                          currentRow++, 0);
        absPassMarkRow = currentRow;
        mLayout.addWidget(
            new QLabel(
                "60% of the maximum number of points that can be achieved:",
                this),
            currentRow++, 0);
    }

    {
        // Separator line
        const auto line = new QFrame(this);
        line->setFrameShape(QFrame::HLine);
        mLayout.addWidget(line, currentRow++, 0, 1, 3);
    }

    {
        // Relative Pass Mark
        mLayout.addWidget(new QLabel("<b>Relative Pass Mark:</b>", this),
                          currentRow++, 0);
        relPassMark0_5Row = currentRow;
        mLayout.addWidget(
            new QLabel(
                "50% of the maximum number of points that can be achieved:",
                this),
            currentRow++, 0);
        relPassMark0_78Row = currentRow;
        mLayout.addWidget(
            new QLabel("78% of the average number of points achieved by all "
                       "students\ntaking the exam for the first time:",
                       this),
            currentRow++, 0);
    }

    {
        // Separator line
        const auto line = new QFrame(this);
        line->setFrameShape(QFrame::HLine);
        mLayout.addWidget(line, currentRow++, 0, 1, 3);
    }

    {
        // Summary
        mLayout.addWidget(&mSummary, currentRow++, 0, 1, 3);
    }

    {
        // Vertical separator line
        // Don't separate summary and supposition
        const auto line = new QFrame(this);
        line->setFrameShape(QFrame::VLine);
        mLayout.addWidget(line, 1, 1, currentRow - 2, 1);
    }

    mLayout.addWidget(&mMaxClosed, maxClosedRow, 2);
    mLayout.addWidget(&mAvgFirstTry, avgFirstTryRow, 2);
    mLayout.addWidget(&mAbsPassMark, absPassMarkRow, 2);
    mLayout.addWidget(&mRelPassMark0_5, relPassMark0_5Row, 2);
    mLayout.addWidget(&mRelPassMark0_78, relPassMark0_78Row, 2);

    connect(&DataManager::getInstance(), &DataManager::gradesChanged, this,
            &CalculationWindow::gradesChanged);
}

void CalculationWindow::gradesChanged() {
    const auto cur = DataManager::getCurrent();
    if (!cur) {
        mMaxClosed.setText("-");
        mAvgFirstTry.setText("-");
        mAbsPassMark.setText("-");
        mRelPassMark0_5.setText("-");
        mRelPassMark0_78.setText("-");
        mSummary.setText("-");
        return;
    }

    // TODO: this has to be changed should we change our stance on attempt nr 0
    const auto hasFirstTry = !cur->mFirstTry.mFirstTryClosedPoints.empty();
    const auto nrDecPlaces =
        nrDecimalPlaces(cur->mUserPointConfig.mGradeBoundaryStepSize);

    const auto changedMaxPoints =
        toPoints(cur->mChangedMaxPointsClosed.mChangedMaxPointBoundary,
                 cur->mUserPointConfig.mGradeBoundaryStepSize);
    mMaxClosed.setText(QString::number(changedMaxPoints.p, 'f', nrDecPlaces));
    if (hasFirstTry) {
        mAvgFirstTry.setText(
            QString::number(cur->mFirstTry.mAvgFirstTryPoints.p));
    } else {
        mAvgFirstTry.setText("None");
    }

    const auto avgPassMark = cur->mClosedPassingLimit.mAbsPassingLimit;
    mAbsPassMark.setText(QString("0.6 * %1 = <b>%2</b>")
                             .arg(changedMaxPoints.p, 0, 'f', nrDecPlaces)
                             .arg(avgPassMark.p));

    // Relative pass mark
    const auto relPassMark0_5 =
        toPoints(cur->mChangedMaxPointsClosed.mChangedMaxPointBoundary,
                 cur->mUserPointConfig.mGradeBoundaryStepSize)
            .p *
        0.5;
    const auto relPassMark0_78 = cur->mFirstTry.mAvgFirstTryPoints.p * 0.78;

    // Change which one is bold depending on which one is bigger
    QString relPassMark0_5Str, relPassMark0_78Str;
    if (relPassMark0_5 >= relPassMark0_78) {
        relPassMark0_5Str = QString("<b>%1</b>").arg(relPassMark0_5);
        relPassMark0_78Str = QString("%1").arg(relPassMark0_78);
    } else {
        relPassMark0_5Str = QString("%1").arg(relPassMark0_5);
        relPassMark0_78Str = QString("<b>%1</b>").arg(relPassMark0_78);
    }

    mRelPassMark0_5.setText(QString("0.5 * %1 = %2")
                                .arg(changedMaxPoints.p, 0, 'f', nrDecPlaces)
                                .arg(relPassMark0_5Str));
    if (!hasFirstTry) {
        mRelPassMark0_78.setText("None");
    } else {
        mRelPassMark0_78.setText(QString("0.78 * %1 = %2")
                                     .arg(cur->mFirstTry.mAvgFirstTryPoints.p)
                                     .arg(relPassMark0_78Str));
    }
    const auto finalRelPassMark =
        hasFirstTry ? qMax(relPassMark0_5, relPassMark0_78) : relPassMark0_5;

    const auto absPassMarkUsed =
        hasFirstTry ? avgPassMark.p < finalRelPassMark : true;
    const auto finalPassMark =
        absPassMarkUsed ? avgPassMark.p : finalRelPassMark;
    const auto finalPassMarkGradeBoundary = GradeBoundary(
        Points{finalPassMark}, cur->mUserPointConfig.mGradeBoundaryStepSize,
        cur->mUserPointConfig.mGradeBoundaryRounding);
    const auto passMarkUsedStr = absPassMarkUsed ? "absolute" : "relative";

    mSummary.setText(
        QString("The (rounded) %1 pass mark of <b>%2 points</b> is decisive.")
            .arg(passMarkUsedStr)
            .arg(toPoints(finalPassMarkGradeBoundary,
                          cur->mUserPointConfig.mGradeBoundaryStepSize)
                     .p,
                 0, 'f', nrDecPlaces));
}
